# The MIT License (MIT)
#
# Copyright (c) 2016-2020 Philippe Proulx <eepp.ca>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re
import os
import sys
import stat
import copy
import shlex
import os.path
import functools
import subprocess
import vlttng.profile
from termcolor import colored
from vlttng.utils import perror
from pathlib import PurePosixPath


def _sq(t):
    return shlex.quote(t)


def _pcmd(cmd):
    print(colored(cmd, attrs=['bold']))


def _psetenv(key, val):
    setenv = 'export {}={}'.format(key.strip(), _sq(str(val)))
    print(colored(setenv, 'grey', attrs=['bold']))


_first_info_done = False


def _comment(msg):
    return '# {}'.format(msg)


def _pinfo(msg):
    global _first_info_done

    if _first_info_done:
        print()
    else:
        _first_info_done = True

    print(colored(_comment(msg), 'blue', attrs=['bold']))


def _pwarn(msg):
    print(colored(_comment('Warning: {}'.format(msg)), 'yellow', attrs=['bold']))


def _patch_env(env, paths):
    # PATH
    path = env.get('PATH', '')
    path = '{}:{}'.format(paths.bin, path)
    env['PATH'] = path

    # CPPFLAGS
    cppflags = env.get('CPPFLAGS', '')
    include_dir = os.path.join(paths.usr, 'include')
    cppflags += ' -I{}'.format(_sq(include_dir))
    env['CPPFLAGS'] = cppflags

    # LDFLAGS
    ldflags = env.get('LDFLAGS', '')
    ldflags += ' -L{}'.format(_sq(paths.lib))
    env['LDFLAGS'] = ldflags

    # LD_LIBRARY_PATH
    ld_library_path = env.get('LD_LIBRARY_PATH', '')
    ld_library_path = '{}:{}'.format(paths.lib, ld_library_path)
    env['LD_LIBRARY_PATH'] = ld_library_path

    # PKG_CONFIG_PATH
    pkg_config_path = env.get('PKG_CONFIG_PATH', '')
    pkg_config_path = '{}:{}'.format(paths.pkgconfig, pkg_config_path)
    env['PKG_CONFIG_PATH'] = pkg_config_path

    # PYTHONPATH
    if os.path.isdir(paths.lib):
        python_roots = []

        for filename in os.listdir(paths.lib):
            if filename.startswith('python'):
                python_root = os.path.join(paths.lib, filename)

                if os.path.isdir(python_root):
                    python_roots.append(python_root)

        site_packages = []

        for python_root in python_roots:
            for filename in os.listdir(python_root):
                if filename.endswith('-packages'):
                    site_package = os.path.join(python_root, filename)

                    if os.path.isdir(site_package):
                        site_packages.append(site_package)

        new_pythonpath = ':'.join(site_packages)
        pythonpath = env.get('PYTHONPATH', '')
        pythonpath = '{}:{}'.format(new_pythonpath, pythonpath)
        env['PYTHONPATH'] = pythonpath


def _get_full_env(env, paths):
    new_env = copy.deepcopy(os.environ)

    try:
        new_env.update(env)
    except:
        perror('Invalid environment in profile')

    _patch_env(new_env, paths)

    return new_env


class _Runner:
    def __init__(self, verbose, hide_export, paths):
        self._verbose = verbose
        self._hide_export = hide_export
        self._cwd = None
        self._env = None
        self._paths = paths

    @property
    def cwd(self):
        return self._cwd

    def _run_line(self, cmd):
        _pcmd(cmd)
        stdio = None if self._verbose else subprocess.DEVNULL
        popen = subprocess.Popen(cmd, stdin=None, stdout=stdio, stderr=stdio,
                                 shell=True, cwd=self._cwd, env=self._env)
        popen.wait()

        if popen.returncode != 0:
            perror('Command exited with status {}'.format(popen.returncode))

    def run(self, cmd):
        if type(cmd) is str:
            self._run_line(cmd)
        elif type(cmd) is list:
            for line in cmd:
                self._run_line(line)

    def cd(self, cwd):
        msg = 'cd {}'.format(_sq(cwd))
        print(colored(msg, 'cyan', attrs=['bold']))
        self._cwd = cwd

    def set_env(self, env):
        self._env = _get_full_env(env, self._paths)

        if not self._hide_export:
            for key in sorted(self._env):
                _psetenv(key, self._env[key])

    def wget(self, url, output_path):
        cmd = 'wget {} -O {}'.format(_sq(url), _sq(output_path))
        self.run(cmd)

    def git_clone(self, clone_url, path):
        cmd = 'git clone {} {}'.format(_sq(clone_url), _sq(path))
        self.run(cmd)

    def git_checkout(self, treeish):
        cmd = 'git checkout {}'.format(_sq(treeish))
        self.run(cmd)

    def mkdir_p(self, path):
        cmd = 'mkdir -v -p {}'.format(_sq(path))
        self.run(cmd)

    def cp_rv(self, src, dst):
        cmd = 'cp -rv {} {}'.format(_sq(src), _sq(dst))
        self.run(cmd)

    def ln_s(self, target, link):
        cmd = 'ln -s {} {}'.format(_sq(target), _sq(link))
        self.run(cmd)

    def tar_x(self, path, output_name):
        cmd = 'tar -xvf {} -C {} --strip-components=1'.format(_sq(path),
                                                              _sq(output_name))
        self.run(cmd)

    def rm_rf(self, path):
        # some basic protection
        no_rm_dirs = (
            os.path.expanduser('~'),
            '/',
            '/bin',
            '/boot',
            '/dev',
            '/etc',
            '/home',
            '/lib',
            '/lib64',
            '/opt',
            '/root',
            '/run',
            '/sbin',
            '/usr',
            '/var'
        )
        no_rm_dirs = [os.path.normpath(d) for d in no_rm_dirs]
        norm_path = os.path.normpath(path)

        if norm_path in no_rm_dirs:
            perror('Not removing protected directory "{}"'.format(norm_path))

        cmd = 'rm -rf {}'.format(_sq(path))
        self.run(cmd)


class _Paths:
    def __init__(self, venv):
        self._venv = venv

    @property
    def venv(self):
        return self._venv

    @property
    def home(self):
        return os.path.join(self._venv, 'home')

    @property
    def usr(self):
        return os.path.join(self._venv, 'usr')

    @property
    def bin(self):
        return os.path.join(self.usr, 'bin')

    @property
    def lib(self):
        return os.path.join(self.usr, 'lib')

    @property
    def pkgconfig(self):
        return os.path.join(self.lib, 'pkgconfig')

    @property
    def include(self):
        return os.path.join(self.usr, 'include')

    @property
    def opt(self):
        return os.path.join(self.usr, 'opt')

    @property
    def src(self):
        return os.path.join(self._venv, 'src')

    @property
    def share(self):
        return os.path.join(self.usr, 'share')

    @property
    def share_java(self):
        return os.path.join(self.share, 'java')

    @property
    def log4j_jar(self):
        return os.path.join(self.share_java, 'log4j.jar')

    def project_src(self, name):
        return os.path.join(self.src, name)


class _ProjectInstructions:
    def __init__(self, project, add_env=None, conf_lines=None,
                 build_lines=None, install_lines=None, uninstall_lines=None):
        self._project = project
        self.add_env = add_env
        self.conf_lines = conf_lines
        self.build_lines = build_lines
        self.install_lines = install_lines
        self.uninstall_lines = uninstall_lines

    @property
    def project(self):
        return self._project


class VEnvCreator:
    _LOG4J_NAME = 'log4j-1.2.17'

    def __init__(self, path, profile, force, verbose, jobs, hide_export):
        self._paths = _Paths(os.path.abspath(path))
        self._runner = _Runner(verbose, hide_export, self._paths)
        self._jobs = jobs
        self._profile = profile
        self._force = force
        self._verbose = verbose
        self._src_paths = {}
        self._project_instructions = {}
        self._create_project_instructions_cbs = {
            'babeltrace': self._create_project_instructions_generic_autotools,
            'babeltrace2': self._create_project_instructions_generic_autotools,
            'elfutils': self._create_project_instructions_generic_autotools,
            'glib': self._create_project_instructions_generic_autotools,
            'libxml2': self._create_project_instructions_generic_autotools,
            'lttng-analyses': self._create_project_instructions_lttng_analyses,
            'lttng-modules': self._create_project_instructions_lttng_modules,
            'lttng-scope': self._create_project_instructions_lttng_scope,
            'lttng-tools': self._create_project_instructions_lttng_tools,
            'lttng-ust': self._create_project_instructions_lttng_ust,
            'popt': self._create_project_instructions_generic_autotools,
            'tracecompass': self._create_project_instructions_tracecompass,
            'urcu': self._create_project_instructions_generic_autotools,
        }
        self._create()

    def _get_make(self):
        return 'make -j{} V=1'.format(self._jobs if self._jobs is not None else '')

    def _check_man_pages(self, name, project):
        if type(project.source) is not vlttng.profile.GitSource:
            return

        disable_man_pages = False

        if '--enable-man-pages=no' not in project.configure and '--disable-man-pages' not in project.configure:
            try:
                subprocess.check_call('asciidoc --version', shell=True,
                                      stdin=subprocess.DEVNULL,
                                      stdout=subprocess.DEVNULL,
                                      stderr=subprocess.DEVNULL,
                                      timeout=1)

                try:
                    subprocess.check_call('xmlto --version', shell=True,
                                          stdin=subprocess.DEVNULL,
                                          stdout=subprocess.DEVNULL,
                                          stderr=subprocess.DEVNULL,
                                          timeout=1)
                except:
                    _pwarn('{} man pages will not be built because vlttng cannot find and execute `xmlto`'.format(name))
                    disable_man_pages = True
            except:
                _pwarn('{} man pages will not be built because vlttng cannot find and execute `asciidoc`'.format(name))
                disable_man_pages = True

        if disable_man_pages:
            project.configure += ' --disable-man-pages'

    def _validate_profile(self):
        def check_dep(project_name, dep_name):
            if project_name in projects and dep_name not in projects:
                _pwarn('The "{}" project will use an external "{}"'.format(project_name, dep_name))

        projects = self._profile.projects
        check_dep('lttng-tools', 'urcu')
        check_dep('lttng-ust', 'urcu')
        check_dep('lttng-analyses', 'babeltrace')
        check_dep('babeltrace', 'glib')
        check_dep('babeltrace2', 'glib')
        check_dep('lttng-tools', 'libxml2')
        check_dep('lttng-tools', 'popt')
        check_dep('babeltrace', 'popt')

        if 'lttng-analyses' in projects:
            if 'babeltrace' in projects and '--enable-python-bindings' not in projects['babeltrace'].configure:
                _pwarn('Configuring "babeltrace" project with "--enable-python-bindings" (caused by "lttng-analyses")')
                configure = projects['babeltrace'].configure
                configure = configure.replace('--disable-python-bindings', '')
                configure = configure.replace('--enable-python-bindings=no', '')
                configure += ' --enable-python-bindings'
                projects['babeltrace'].configure = configure

        if 'babeltrace' in projects:
            project = projects['babeltrace']

            if '--enable-debug-info' in project.configure or '--disable-debug-info' not in project.configure:
                check_dep('babeltrace', 'elfutils')

        if 'babeltrace2' in projects:
            project = projects['babeltrace2']

            if '--enable-debug-info' in project.configure or '--disable-debug-info' not in project.configure:
                check_dep('babeltrace2', 'elfutils')

        if 'lttng-tools' in projects:
            self._check_man_pages('LTTng-tools', projects['lttng-tools'])

        if 'lttng-ust' in projects:
            self._check_man_pages('LTTng-UST', projects['lttng-ust'])

    def _create_project_instructions_lttng_tools(self, project):
        lttng_ust_opts = (
            '--with-lttng-ust-prefix',
            '--without-lttng-ust',
            '--disable-lttng-ust',
            '--enable-lttng-ust',
        )

        for ust_opt in lttng_ust_opts:
            if ust_opt in project.configure:
                fmt = 'I would not pass the {} configure option if I were you: they are handled by vlttng'
                msg = fmt.format(ust_opt)
                _pwarn(msg)

        add_args = ''

        if 'lttng-ust' not in self._profile.projects:
            # LTTng-tools prior to v2.8 uses a different flag to turn
            # off LTTng-UST support. We add both to the configure script
            # arguments: the unsupported one will be ignored.
            add_args = '--without-lttng-ust --disable-lttng-ust'

        add_args += ' --disable-kmod --without-kmod'

        return self._create_project_instructions_generic_autotools(project, add_args)

    def _create_project_instructions_lttng_ust(self, project):
        instructions = self._create_project_instructions_generic_autotools(project)
        instructions.add_env = {
            'CLASSPATH': self._paths.log4j_jar,
        }

        return instructions

    def _create_project_instructions_lttng_modules(self, project):
        build_lines = [
            self._get_make(),
        ]
        sq_install_path = _sq(self._paths.usr)
        install_lines = [
            'make modules_install INSTALL_MOD_PATH={}'.format(sq_install_path),
            'depmod --all --basedir={}'.format(sq_install_path),
        ]

        return _ProjectInstructions(project, build_lines=build_lines,
                                    install_lines=install_lines)

    def _create_project_instructions_lttng_analyses(self, project):
        build_lines = [
            './setup.py build',
        ]
        install_lines = [
            './setup.py install --prefix={}'.format(_sq(self._paths.usr))
        ]

        return _ProjectInstructions(project, build_lines=build_lines,
                                    install_lines=install_lines)

    def _create_project_instructions_tracecompass(self, project):
        dst = os.path.join(self._paths.opt, 'tracecompass')
        install_lines = []

        if type(project.source) is vlttng.profile.HttpFtpSource:
            src = self._paths.project_src('tracecompass')
        else:
            install_lines = [
                'mvn clean install -Dmaven.test.skip=true',
            ]
            src = os.path.join('rcp',
                               'org.eclipse.tracecompass.rcp.product',
                               'target',
                               'products',
                               'org.eclipse.tracecompass.rcp',
                               'linux',
                               'gtk',
                               'x86_64',
                               'trace-compass')

        link = os.path.join(self._paths.bin, 'tracecompass')
        install_lines += [
            'cp -rv {} {}'.format(_sq(src), _sq(dst)),
            'ln -s {} {}'.format(_sq(os.path.join(dst, 'tracecompass')),
                                 _sq(link)),
        ]

        return _ProjectInstructions(project, install_lines=install_lines)

    def _create_project_instructions_lttng_scope(self, project):
        jar_dst = os.path.join(self._paths.opt, 'lttng-scope.jar')

        if type(project.source) is vlttng.profile.HttpFtpSource:
            jar_src = os.path.join(self._paths.src, 'lttng-scope.jar')
            install_lines = [
                'cp -v {} {}'.format(_sq(jar_src), _sq(jar_dst)),
            ]
        else:
            jar_dir = os.path.join('lttng-scope', 'target')
            install_lines = [
                'mvn clean install -Dmaven.test.skip=true -DskipTests',
                'cp -v {}/*with-dependencies.jar {}'.format(_sq(jar_dir),
                                                             _sq(jar_dst)),
            ]

        return _ProjectInstructions(project, install_lines=install_lines)

    def _create_project_instructions_generic_autotools(self, project, add_conf_args=None):
        conf_lines = []
        build_lines = [self._get_make()]
        install_lines = ['make install']
        uninstall_lines = ['make uninstall']

        # bootstrap?
        project_src = self._paths.project_src(project.name)

        if not os.path.isfile(os.path.join(project_src, 'configure')):
            for f in ('bootstrap', 'bootstrap.sh', 'autogen', 'autogen.sh',):
                bootstrap = os.path.join(project_src, f)

                if os.path.isfile(bootstrap):
                    conf_lines.append('./{}'.format(f))
                    break

        # configure
        conf_args = project.configure

        if '--prefix' in project.configure:
            fmt = 'Project "{}": I would not pass the --prefix configure option if I were you: it is handled by vlttng'
            _pwarn(fmt.format(project.name))

        if add_conf_args is not None:
            conf_args += ' ' + add_conf_args

        conf_line = './configure --prefix={} {}'.format(_sq(self._paths.usr),
                                                        conf_args)
        conf_lines.append(conf_line)

        return _ProjectInstructions(project, conf_lines=conf_lines,
                                    build_lines=build_lines,
                                    install_lines=install_lines,
                                    uninstall_lines=uninstall_lines)

    def _create_project_instructions(self):
        for name, project in self._profile.projects.items():
            self._project_instructions[name] = self._create_project_instructions_cbs[name](project)

    def _create(self):
        self._validate_profile()

        _pinfo('Create LTTng virtual environment')

        # create virtual environment directory
        if os.path.exists(self._paths.venv):
            if self._force:
                _pwarn('Virtual environment path "{}" exists: removing directory'.format(self._paths.venv))
                self._runner.rm_rf(self._paths.venv)
            else:
                perror('Virtual environment path "{}" exists (use --force to overwrite)'.format(self._paths.venv))

        self._runner.mkdir_p(self._paths.venv)
        self._runner.mkdir_p(self._paths.home)
        self._runner.mkdir_p(self._paths.bin)
        self._runner.mkdir_p(self._paths.lib)
        self._runner.mkdir_p(self._paths.include)
        self._runner.mkdir_p(self._paths.opt)
        self._runner.mkdir_p(self._paths.share_java)

        # fetch sources and extract/checkout
        _pinfo('Fetch sources')
        self._fetch_sources()

        # create build instructions for projects to build
        self._create_project_instructions()

        # build projects in this order
        self._build_project('urcu')
        self._build_project('popt')
        self._build_lttng_ust()
        self._build_project('libxml2')
        self._build_project('lttng-tools')
        self._build_project('lttng-modules')
        self._build_project('glib')
        self._build_project('elfutils')
        self._build_project('babeltrace')
        self._build_project('babeltrace2')
        self._build_project('lttng-analyses')
        self._build_project('tracecompass')
        self._build_project('lttng-scope')

        # create activate script
        self._create_activate()

    def _create_activate(self):
        from vlttng.activate_template import activate_template

        env_items = []
        unenv_items = []
        env = copy.deepcopy(self._profile.virt_env)
        _patch_env(env, self._paths)

        # the activation template explicitly defines those environment
        # variables, so they must not be overridden by the user or by us.
        rm_keys = (
            'VLTTNG',
            'PATH',
            'CPPFLAGS',
            'LDFLAGS',
            'LD_LIBRARY_PATH',
            'MANPATH',
            'PKG_CONFIG_PATH',
            'PYTHONPATH',
            'LTTNG_HOME',
            'PS1',
            'MODPROBE_OPTIONS',
        )

        for key in rm_keys:
            if key in env:
                del env[key]

        for key, val in env.items():
            key = key.strip()
            env_items.append('vlttng-save-env {}'.format(key))
            setenv = 'export {}={}'.format(key, _sq(str(val)))
            env_items.append(setenv)
            unenv_items.append('    vlttng-restore-env {}'.format(key))

        env_lines = '\n'.join(env_items)
        unenv_lines = '\n'.join(unenv_items)
        lttng_modules_src_path = self._src_paths.get('lttng-modules', '')
        has_modules = '1' if 'lttng-modules' in self._profile.projects else '0'
        has_lttng_scope = '1' if 'lttng-scope' in self._profile.projects else '0'
        has_java = '0'

        if 'lttng-ust' in self._profile.projects:
            configure = self._profile.projects['lttng-ust'].configure

            if '--enable-java-agent' in configure:
                has_java = '1'

        activate = activate_template.format(venv_path=_sq(self._paths.venv),
                                            has_modules=has_modules,
                                            has_java=has_java,
                                            has_lttng_scope=has_lttng_scope,
                                            env=env_lines,
                                            unenv=unenv_lines)
        activate_path = os.path.join(self._paths.venv, 'activate')
        _pinfo('Create activation script "{}"'.format(activate_path))

        with open(activate_path, 'w') as f:
            f.write(activate)

    def _fetch_sources(self):
        self._runner.mkdir_p(self._paths.src)
        self._runner.cd(self._paths.src)

        for project in self._profile.projects.values():
            source = project.source
            src_path = None

            if type(source) is vlttng.profile.HttpFtpSource:
                # download
                posix_path = PurePosixPath(source.url)

                if project.name == 'lttng-scope':
                    src_path = self._paths.src
                    filename = 'lttng-scope.jar'
                else:
                    src_path = project.name
                    filename = posix_path.name

                self._runner.wget(source.url, filename)

                # extract
                if not filename.endswith('.jar'):
                    self._runner.mkdir_p(self._paths.project_src(project.name))
                    self._runner.tar_x(filename, project.name)
            elif type(source) is vlttng.profile.GitSource:
                src_path = project.name

                # clone
                self._runner.git_clone(source.clone_url, project.name)

                # checkout
                self._runner.cd(self._paths.project_src(project.name))
                self._runner.git_checkout(source.checkout)
                self._runner.cd(self._paths.src)

            # keep where the source of this project is
            if src_path is not None:
                src_path = self._paths.project_src(src_path)
                self._src_paths[project.name] = src_path

    def _build_lttng_ust(self):
        project = self._profile.projects.get('lttng-ust')

        if project is None:
            return

        if '--enable-java-agent-all' in project.configure or '--enable-java-agent-log4j' in project.configure:
            # get Apache log4j 1.2
            log4j_name = 'log4j-1.2.17'
            log4j_tarball = '{}.tar.gz'.format(log4j_name)
            log4j_jar = '{}.jar'.format(log4j_name)
            _pinfo('Download Apache log4j')
            self._runner.cd(self._paths.src)
            self._runner.wget('http://apache.mirror.gtcomm.net/logging/log4j/1.2.17/{}'.format(log4j_tarball),
                              log4j_tarball)

            # extract
            self._runner.mkdir_p(log4j_name)
            self._runner.tar_x(log4j_tarball, log4j_name)

            # install
            self._runner.cp_rv(os.path.join(log4j_name, log4j_jar),
                               self._paths.log4j_jar)

        self._build_project('lttng-ust')

    def _get_build_env_from_instructions(self, instructions):
        build_env = copy.deepcopy(self._profile.build_env)
        build_env.update(instructions.project.build_env)

        if instructions.add_env is not None:
            build_env.update(instructions.add_env)

        return build_env

    def _build_project(self, name):
        instructions = self._project_instructions.get(name)

        if instructions is None:
            return

        build_env = self._get_build_env_from_instructions(instructions)
        self._runner.set_env(build_env)
        self._runner.cd(self._src_paths[instructions.project.name])

        if instructions.conf_lines is not None:
            _pinfo('Configure {}'.format(name))
            self._runner.run(instructions.conf_lines)

        if instructions.build_lines is not None:
            _pinfo('Build {}'.format(name))
            self._runner.run(instructions.build_lines)

        if instructions.install_lines is not None:
            _pinfo('Install {}'.format(name))
            self._runner.run(instructions.install_lines)

        self._create_scripts(instructions)

    def _create_executable_script(self, script_name, content):
        script_path = os.path.join(self._paths.venv,
                                   '{}.bash'.format(script_name))

        with open(script_path, 'w') as f:
            f.write(content)

        st = os.stat(script_path)
        os.chmod(script_path, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

    def _create_conf_script(self, instructions, name, exports, src_path):
        from vlttng.conf_template import conf_template as tmpl

        conf_lines = ''

        if instructions.conf_lines is not None:
            conf_lines = '\n'.join(instructions.conf_lines)

        conf = tmpl.format(name=name, src_path=_sq(src_path),
                           conf_lines=conf_lines, exports=exports)
        self._create_executable_script('conf-{}'.format(name), conf)

    def _create_build_script(self, instructions, name, exports, src_path):
        from vlttng.build_template import build_template as tmpl

        build_lines = ''

        if instructions.build_lines is not None:
            build_lines = '\n'.join(instructions.build_lines)

        build = tmpl.format(name=name, src_path=_sq(src_path),
                            build_lines=build_lines, exports=exports)
        self._create_executable_script('build-{}'.format(name), build)

    def _create_install_script(self, instructions, name, exports, src_path):
        from vlttng.install_template import install_template as tmpl

        install_lines = ''

        if instructions.install_lines is not None:
            install_lines = '\n'.join(instructions.install_lines)

        install = tmpl.format(name=name, src_path=_sq(src_path),
                              install_lines=install_lines, exports=exports)
        self._create_executable_script('install-{}'.format(name), install)

    def _create_update_script(self, instructions, name, exports, src_path):
        from vlttng.update_template import update_template as tmpl

        gitref = instructions.project.source.checkout
        uninstall_lines = ''

        if instructions.uninstall_lines is not None:
            uninstall_lines = '\n'.join(instructions.uninstall_lines)

        update = tmpl.format(name=name, src_path=_sq(src_path),
                             uninstall_lines=uninstall_lines, gitref=gitref,
                             exports=exports)
        self._create_executable_script('update-{}'.format(name), update)

    def _create_scripts(self, instructions):
        name = instructions.project.name
        src_path = self._src_paths[name]
        build_env = self._get_build_env_from_instructions(instructions)
        build_env = _get_full_env(build_env, self._paths)
        export_lines = []

        for key in sorted(build_env):
            value = build_env[key]
            export_lines.append('export {}={}'.format(key, _sq(value)))

        exports = '\n'.join(export_lines)

        # always generate those
        self._create_conf_script(instructions, name, exports, src_path)
        self._create_build_script(instructions, name, exports, src_path)
        self._create_install_script(instructions, name, exports, src_path)

        if type(instructions.project.source) is vlttng.profile.GitSource:
            # only generate update script if it's a Git source
            self._create_update_script(instructions, name, exports, src_path)



# The MIT License (MIT)
#
# Copyright (c) 2016 Philippe Proulx <eepp.ca>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from vlttng.utils import perror
import vlttng.profile
import pkg_resources
import vlttng.venv
import argparse
import platform
import os.path
import vlttng
import re
import os


def _parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('-f', '--force', action='store_true',
                    help='force the virtual environment creation')
    ap.add_argument('--hide-export', action='store_true',
                    help='hide export lines')
    ap.add_argument('-i', '--ignore-project', metavar='PROJECT',
                    action='append',
                    help='ignore project PROJECT (may be repeated)')
    ap.add_argument('-j', '--jobs', nargs='?', const=None, metavar='JOBS',
                    action='store', type=int,
                    default=1, help='number of make jobs to run simultaneously')
    ap.add_argument('-l', '--list-default-profiles', action='store_true',
                    help='list default profile names and exit')
    ap.add_argument('-o', '--override', metavar='PROP',
                    action='append',
                    help='override property in the effective profile (may be repeated)')
    ap.add_argument('-p', '--profile', metavar='PROFILE', action='append',
                    help='profile name or path (may be repeated to patch)')
    ap.add_argument('-v', '--verbose', action='store_true',
                    help='verbose output')
    ap.add_argument('-V', '--version', action='version',
                    version='%(prog)s {}'.format(vlttng.__version__))
    ap.add_argument('path', metavar='PATH', action='store', nargs='*',
                    help='virtual environment path')

    # parse args
    args = ap.parse_args()

    if not args.list_default_profiles and len(args.path) != 1:
        perror('Exactly one virtual environment path must be provided')

    if len(args.path) > 0:
        args.path = args.path[0]

    if args.ignore_project is None:
        args.ignore_project = []

    if args.profile is None:
        args.profile = []

    if args.override is None:
        args.override = []

    return args


def _find_profile(profile_name):
    try:
        filename = os.path.join(vlttng._PROFILES_DIRNAME,
                                '{}.yml'.format(profile_name))
        filename = pkg_resources.resource_filename(__name__, filename)

        if not os.path.isfile(filename):
            filename = profile_name
    except:
        filename = profile_name

    if not os.path.isfile(filename):
        perror('Cannot find profile "{}"'.format(profile_name))

    return filename


def _create_overrides(override_args):
    overrides = []

    try:
        for arg in override_args:
            arg = arg.strip()

            # replace
            m = re.match(r'^([a-zA-Z0-9-_]+(?:\.[a-zA-Z0-9-_]+)*)=(.+)$', arg)

            if m:
                path = m.group(1).split('.')
                rep = m.group(2)
                op = vlttng.profile.Override.OP_REPLACE
                overrides.append(vlttng.profile.Override(path, op, rep))
                continue

            # append
            m = re.match(r'^([a-zA-Z0-9-_]+(?:\.[a-zA-Z0-9-_]+)*)\+=(.+)$', arg)

            if m:
                path = m.group(1).split('.')
                rep = m.group(2)
                op = vlttng.profile.Override.OP_APPEND
                overrides.append(vlttng.profile.Override(path, op, rep))
                continue

            # remove
            m = re.match(r'^!([a-zA-Z0-9-_]+(?:\.[a-zA-Z0-9-_]+)*)$', arg)

            if m:
                path = m.group(1).split('.')
                op = vlttng.profile.Override.OP_REMOVE
                overrides.append(vlttng.profile.Override(path, op, None))
                continue

            perror('Malformed override option: "{}"'.format(arg))
    except vlttng.profile.InvalidOverride as e:
        perror('In override option: "{}": {}'.format(arg, e))

    return overrides


def _create_profile(profile_names, ignored_projects, override_args, verbose):
    filenames = []

    try:
        overrides = _create_overrides(override_args)
    except Exception as e:
        perror('Cannot parse overrides: {}'.format(e))

    for profile_name in profile_names:
        filenames.append(_find_profile(profile_name))

    try:
        profile = vlttng.profile.from_yaml_files(filenames, ignored_projects,
                                                 overrides, verbose)
    except vlttng.profile.ParseError:
        perror('Malformed YAML profile')
    except vlttng.profile.UnknownSourceFormat as e:
        perror('Unknown source format: "{}"'.format(e.source))
    except vlttng.profile.InvalidProfile as e:
        perror('Invalid profile: {}'.format(e))
    except:
        perror('Cannot read one of the specified profiles')

    return profile


def _register_sigint():
    if platform.system() == 'Linux':
        def handler(signal, frame):
            perror('Cancelled by user: virtual environment is incomplete')

        import signal
        signal.signal(signal.SIGINT, handler)


def _list_default_profiles():
    dirname = pkg_resources.resource_filename(__name__,
                                              vlttng._PROFILES_DIRNAME)

    for filename in sorted(os.listdir(dirname)):
        if filename.endswith('.yml'):
            print(filename[:-4])


def run():
    _register_sigint()
    args = _parse_args()

    if args.list_default_profiles:
        _list_default_profiles()
        return 0

    profile = _create_profile(args.profile, args.ignore_project, args.override,
                              args.verbose)

    try:
        vlttng.venv.VEnvCreator(args.path, profile, args.force, args.verbose,
                                args.jobs, args.hide_export)
    except Exception as e:
        perror('Unexpected error: {}'.format(e))

    return 0



# The MIT License (MIT)
#
# Copyright (c) 2016-2020 Philippe Proulx <eepp.ca>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from collections import namedtuple
from vlttng.utils import perror
from termcolor import colored
import pkg_resources
import argparse
import platform
import vlttng
import shlex
import enum
import os
import re


def _parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('-V', '--version', action='version',
                    version='%(prog)s {}'.format(vlttng.__version__))

    # parse args
    args = ap.parse_args()

    return args


def _register_sigint():
    if platform.system() == 'Linux':
        def handler(signal, frame):
            perror('Cancelled by user')

        import signal
        signal.signal(signal.SIGINT, handler)


def _bold(s):
    return colored(s, attrs=['bold'])


def _cquestion(s):
    return colored(s, 'cyan', attrs=['bold'])


def _cchoice(s):
    return _bold(s)


def _cprompt(p):
    return colored(p, 'yellow', attrs=['bold'])


def _cerror(e):
    return colored(e, 'red', attrs=['bold'])


@enum.unique
class _WizardState(enum.Enum):
    ASK_PROJECTS = 1
    ASK_MASTER = 2
    ASK_VERSIONS = 3
    ASK_FEATURE = 4
    ASK_BT_PYTHON = 5
    ASK_BT2_PYTHON = 6
    ASK_LTTNG_TOOLS_PYTHON = 7
    ASK_LTTNG_UST_JUL_AGENT = 8
    ASK_LTTNG_UST_LOG4J_AGENT = 9
    ASK_LTTNG_UST_PYTHON_AGENT = 10
    ASK_PYTHON_INTERPRETER = 11
    ASK_PATH = 12
    END = 13


class _Wizard:
    def __init__(self):
        self._profiles = []
        self._projects = []
        self._projects_versions = {}
        self._python_interpreter = None
        self._path = None
        self._state_handlers = {
            _WizardState.ASK_PROJECTS: self._state_ask_projects,
            _WizardState.ASK_MASTER: self._state_ask_all_master,
            _WizardState.ASK_VERSIONS: self._state_ask_versions,
            _WizardState.ASK_FEATURE: self._state_ask_feature,
            _WizardState.ASK_BT_PYTHON: self._state_ask_bt_python,
            _WizardState.ASK_BT2_PYTHON: self._state_ask_bt2_python,
            _WizardState.ASK_LTTNG_TOOLS_PYTHON: self._state_ask_lttng_tools_python,
            _WizardState.ASK_LTTNG_UST_JUL_AGENT: self._state_ask_lttng_ust_jul_agent,
            _WizardState.ASK_LTTNG_UST_LOG4J_AGENT: self._state_ask_lttng_ust_log4j_agent,
            _WizardState.ASK_LTTNG_UST_PYTHON_AGENT: self._state_ask_lttng_ust_python_agent,
            _WizardState.ASK_PYTHON_INTERPRETER: self._state_ask_python_interpreter,
            _WizardState.ASK_PATH: self._state_ask_path,
        }
        self._project_name_to_title = {
            'babeltrace': ('Babeltrace 1',),
            'babeltrace2': ('Babeltrace 2',),
            'elfutils': ('elfutils', 'optional dependency of Babeltrace (1 and 2)'),
            'glib': ('GLib', 'dependency of Babeltrace (1 and 2)'),
            'libxml2': ('libxml2', 'dependency of LTTng-tools'),
            'lttng-analyses': ('LTTng analyses',),
            'lttng-modules': ('LTTng-modules',),
            'lttng-scope': ('LTTng Scope',),
            'lttng-tools': ('LTTng-tools',),
            'lttng-ust': ('LTTng-UST',),
            'popt': ('popt', 'dependency of Babeltrace 1 and LTTng-tools'),
            'tracecompass': ('Trace Compass',),
            'urcu': ('Userspace RCU', 'dependency of LTTng-tools and LTTng-UST'),
        }
        self._project_name_to_versions = {
            'babeltrace': [],
            'babeltrace2': [],
            'elfutils': [],
            'glib': [],
            'libxml2': [],
            'lttng-analyses': [],
            'lttng-modules': [],
            'lttng-scope': [],
            'lttng-tools': [],
            'lttng-ust': [],
            'popt': [],
            'tracecompass': [],
            'urcu': [],
        }

        dirname = pkg_resources.resource_filename(__name__,
                                                  vlttng._PROFILES_DIRNAME)

        for filename in sorted(os.listdir(dirname)):
            if filename.endswith('.yml'):
                profile = filename[:-4]

                if not re.match(r'^.+-\d+(\.\d+)*$', profile):
                    continue

                for project_name in self._project_name_to_versions:
                    if profile.startswith(project_name + '-'):
                        profile_suffix = profile[len(project_name) + 1:]
                        self._project_name_to_versions[project_name].append(profile_suffix)

    def _handle_state(self):
        self._state_handlers[self._state]()

    def _get_cmd_line_args(self):
        args = ['vlttng']

        for profile in self._profiles:
            args += ['-p', profile]


        if self._python_interpreter is not None:
            args += ['-o', 'build-env.PYTHON={}'.format(self._python_interpreter)]
            args += ['-o', 'build-env.PYTHON_CONFIG={}-config'.format(self._python_interpreter)]

        args.append(shlex.quote(self._path))

        return args

    def start(self):
        self._state = _WizardState.ASK_PROJECTS
        print('''Welcome to the vlttng quickstart!

This questionnaire will help you create a basic vlttng profile to build
the projects and features you need.
''')

        while self._state != _WizardState.END:
            self._handle_state()

        args = self._get_cmd_line_args()

        print(_bold('- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -'))
        print()
        print('Your command line is:')
        print()
        print(_bold(' '.join(args)))
        print()
        print('Note: You can add the {} argument to use N make jobs.'.format(_bold('-jN')))
        print()
        print(_cquestion('Would you like to run this command line?'))
        print()

        if self._get_yes_no(True):
            print()
            print(_cquestion('How many make jobs would you like?'))
            print()

            import multiprocessing

            prompt_ext = ' [{}]'.format(multiprocessing.cpu_count())

            while True:
                answer = self._input(prompt_ext, '')
                print()

                if answer == '':
                    ianswer = multiprocessing.cpu_count()
                    break

                try:
                    ianswer = int(answer.strip())
                except:
                    continue

                if ianswer < 1:
                    continue

                break

            args.append('-j{}'.format(ianswer))

            import subprocess

            try:
                subprocess.run(args)
            except:
                pass

    def _state_ask_projects(self):
        print(_cquestion('Which projects would you like to build?'))
        print()
        choices = (
            self._get_project_title('babeltrace'),
            self._get_project_title('babeltrace2'),
            self._get_project_title('elfutils'),
            self._get_project_title('glib'),
            self._get_project_title('libxml2'),
            self._get_project_title('lttng-analyses'),
            self._get_project_title('lttng-modules'),
            self._get_project_title('lttng-scope'),
            self._get_project_title('lttng-tools'),
            self._get_project_title('lttng-ust'),
            self._get_project_title('popt'),
            self._get_project_title('tracecompass'),
            self._get_project_title('urcu'),
        )
        choice_projects = (
            'babeltrace',
            'babeltrace2',
            'elfutils',
            'glib',
            'libxml2',
            'lttng-analyses',
            'lttng-modules',
            'lttng-scope',
            'lttng-tools',
            'lttng-ust',
            'popt',
            'tracecompass',
            'urcu',
        )
        self._pchoices(choices)
        self._pmultiple_choices_info()
        choices = self._get_choices(len(choices))
        print()

        if choices == 'a':
            self._projects = [project for project in choice_projects]
        else:
            for choice in choices:
                self._projects.append(choice_projects[choice - 1])

        self._state = _WizardState.ASK_MASTER

    def _state_ask_all_master(self):
        print(_cquestion('Which projects would you like to be built from their Git master version?'))
        print()
        choices = []
        choice_projects = []

        for project in self._projects:
            if project not in ('babeltrace', 'elfutils', 'popt'):
                choices.append(self._get_project_title(project))
                choice_projects.append(project)

        if len(choice_projects) == 0:
            self._state = _WizardState.ASK_VERSIONS
            return

        self._pchoices(choices)
        self._pmultiple_choices_info(with_none=True)
        choices = self._get_choices(len(choices), True)
        print()

        if choices == 'a':
            self._projects_versions = {p: 'master' for p in choice_projects}
        else:
            for choice in choices:
                self._projects_versions[choice_projects[choice - 1]] = 'master'

        self._state = _WizardState.ASK_VERSIONS

    def _state_ask_versions(self):
        projects = []

        for project in self._projects:
            if project not in self._projects_versions:
                projects.append(project)

        for project in projects:
            title = self._get_project_title(project)
            question = 'Which version of {} do you want to build?'.format(title[0])
            print(_cquestion(question))
            print()
            choices = []
            versions = self._project_name_to_versions[project]

            for version in versions:
                choices.append((version,))

            self._pchoices(choices)
            print()
            choice = self._get_choice(len(choices))
            print()

            version = versions[choice - 1]
            self._projects_versions[project] = version

        for project, version in self._projects_versions.items():
            self._profiles.append('{}-{}'.format(project, version))

        self._state = _WizardState.ASK_FEATURE

    def _state_ask_feature(self):
        self._state = _WizardState.ASK_BT_PYTHON

    def _state_ask_bt_python(self):
        next_state = _WizardState.ASK_BT2_PYTHON

        if 'babeltrace' not in self._projects:
            self._state = next_state
            return

        if 'lttng-analyses' in self._projects:
            self._profiles.append('babeltrace-python')
            self._state = next_state
            return

        print(_cquestion('Would you like to build the Babeltrace 1 Python bindings?'))
        print()

        if self._get_yes_no(False):
            self._profiles.append('babeltrace-python')

        print()
        self._state = next_state

    def _state_ask_bt2_python(self):
        next_state = _WizardState.ASK_LTTNG_TOOLS_PYTHON

        if 'babeltrace2' not in self._projects:
            self._state = next_state
            return

        print(_cquestion('Would you like to build the Babeltrace 2 Python bindings?'))
        print()

        if self._get_yes_no(False):
            self._profiles.append('babeltrace2-python')

        print()
        self._state = next_state

    def _state_ask_lttng_tools_python(self):
        if 'lttng-tools' not in self._projects:
            self._state = _WizardState.ASK_LTTNG_UST_JUL_AGENT
            return

        print(_cquestion('Would you like to build the LTTng-tools Python bindings?'))
        print()

        if self._get_yes_no(False):
            self._profiles.append('lttng-tools-python')

        print()
        self._state = _WizardState.ASK_LTTNG_UST_JUL_AGENT

    def _state_ask_lttng_ust_jul_agent(self):
        if 'lttng-ust' not in self._projects:
            self._state = _WizardState.ASK_LTTNG_UST_LOG4J_AGENT
            return

        print(_cquestion('Would you like to build the LTTng-UST java.util.logging agent?'))
        print()

        if self._get_yes_no(False):
            self._profiles.append('lttng-ust-jul-agent')

        print()
        self._state = _WizardState.ASK_LTTNG_UST_LOG4J_AGENT

    def _state_ask_lttng_ust_log4j_agent(self):
        if 'lttng-ust' not in self._projects:
            self._state = _WizardState.ASK_LTTNG_UST_PYTHON_AGENT
            return

        print(_cquestion('Would you like to build the LTTng-UST log4j agent?'))
        print()

        if self._get_yes_no(False):
            self._profiles.append('lttng-ust-log4j-agent')

        print()
        self._state = _WizardState.ASK_LTTNG_UST_PYTHON_AGENT

    def _state_ask_lttng_ust_python_agent(self):
        if 'lttng-ust' not in self._projects:
            self._state = _WizardState.ASK_PYTHON_INTERPRETER
            return

        print(_cquestion('Would you like to build the LTTng-UST Python agent?'))
        print()

        if self._get_yes_no(False):
            self._profiles.append('lttng-ust-python-agent')

        print()
        self._state = _WizardState.ASK_PYTHON_INTERPRETER

    def _state_ask_python_interpreter(self):
        found_python = False

        for profile in self._profiles:
            if 'python' in profile:
                found_python = True
                break

        if not found_python:
            self._state = _WizardState.ASK_PATH
            return

        print(_cquestion("What's your preferred Python interpreter?"))
        print()
        choices = (
            ('Python 2',),
            ('Python 3',),
        )

        self._pchoices(choices)
        print()
        choice = self._get_choice(len(choices))
        print()

        if choice == 1:
            self._python_interpreter = 'python2'
        else:
            self._python_interpreter = 'python3'

        self._state = _WizardState.ASK_PATH

    def _state_ask_path(self):
        print(_cquestion('Where would you like to create the virtual environment?'))
        print()
        answer = self._input(' [virt-lttng]', suffix='')
        print()

        if answer == '':
            answer = 'virt-lttng'

        self._path = answer
        self._state = _WizardState.END

    def _get_project_title(self, name):
        return self._project_name_to_title[name]

    def _pmultiple_choices_info(self, with_none=False):
        print()
        print('Enter a space-delimited list of choices,', end='')

        if with_none:
            print(' {} to select all,'.format(_bold('a')))
            print('or {} to select none.'.format(_bold('n')))
        else:
            print(' or {} to select all.'.format(_bold('a')))

        print()

    def _pchoices(self, choices):
        index_len = len(str(len(choices)))

        for index, choice in enumerate(choices):
            sinfo = ''

            if len(choice) > 1:
                sinfo = ' ({})'.format(choice[1])

            sindex = str(index + 1) + '.'
            sindex = sindex.rjust(index_len + 1)
            print('{} {}{}'.format(sindex, _cchoice(choice[0]), sinfo))

    def _perror(self, error):
        print(_cerror(error))

    def _pinvalid_choices(self, resp=None, suffix='s'):
        print()

        if resp is None:
            print()
            self._perror('Invalid choice{}'.format(suffix))
        else:
            self._perror('Invalid choice{}: {}'.format(suffix, resp))

        print()

    def _raw_choice_to_int(self, raw_choice, choice_max):
        try:
            choice = int(raw_choice)
        except:
            return

        if choice < 1 or choice > choice_max:
            return

        return choice

    def _try_get_choices(self, choice_max, all_default):
        prompt_ext = ''

        if all_default:
            prompt_ext = ' ({} for none) [{}]'.format(_bold('n'), _bold('a'))

        resp = self._input(prompt_ext)

        if resp == '':
            if all_default:
                return 'a'

            self._pinvalid_choices()
            return

        if resp == 'n':
            return set()

        raw_choices = re.split(r'\s+', resp)
        choices = set()

        for raw_choice in raw_choices:
            if raw_choice == 'a':
                return 'a'

            choice = self._raw_choice_to_int(raw_choice, choice_max)

            if choice is None:
                self._pinvalid_choices(resp)
                return

            choices.add(choice)

        return sorted(list(choices))

    def _get_choices(self, choice_max, all_default=False):
        while True:
            choices = self._try_get_choices(choice_max, all_default)

            if choices is not None:
                return choices

    def _try_get_choice(self, choice_max):
        resp = self._input(suffix='')

        if resp == '':
            self._pinvalid_choices(None, '')
            return

        choice = self._raw_choice_to_int(resp, choice_max)

        if choice is None:
            self._pinvalid_choices(resp, '')
            return

        return choice

    def _get_choice(self, choice_max):
        while True:
            choice = self._try_get_choice(choice_max)

            if choice is not None:
                return choice

    def _get_yes_no(self, default):
        default_yn = 'y' if default else 'n'
        prompt_ext = ' [{}]'.format(default_yn)

        while True:
            resp = self._input(prompt_ext, '')

            if resp == '':
                return default
            elif resp == 'y':
                return True
            elif resp == 'n':
                return False

            self._pinvalid_choices(resp, '')

    def _input(self, prompt_ext='', suffix=None):
        prompt = '{}{} '.format(_cprompt('==>'), prompt_ext)

        while True:
            try:
                resp = input(prompt)
            except EOFError:
                self._pinvalid_choices(None, suffix)
                continue

            return resp.strip()


def run():
    _register_sigint()
    args = _parse_args()
    wiz = _Wizard()
    wiz.start()

    return 0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vlttng']

package_data = \
{'': ['*'], 'vlttng': ['profiles/*']}

install_requires = \
['pyyaml>=5.1,<6.0', 'setuptools', 'termcolor>=1.1,<2.0']

entry_points = \
{'console_scripts': ['vlttng = vlttng.vlttng_cli:run',
                     'vlttng-quick = vlttng.vlttng_quick_cli:run']}

setup_kwargs = {
    'name': 'vlttng',
    'version': '0.10.7',
    'description': 'Generator of LTTng virtual environment',
    'long_description': None,
    'author': 'Philippe Proulx',
    'author_email': 'eeppeliteloop@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/eepp/vlttng/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
